%option noyywrap
%option nounput
%option noyylineno
%{
/* Copyright 1995 Yggdrasil Computing, Incorporated.
   Written by Adam J. Richter (adam@yggdrasil.com).
   This file may be copied under the terms and conditions of version 2
   of the GNU General Public License, as published by the Free Software
   Foundation (Cambridge, Massachusetts, USA).
*/

#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/param.h>
#include <time.h>
#include <regex.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <getopt.h>
#include "file.h"

#ifndef ZCAT
#define ZCAT "zcat"
#endif

#ifndef BZCAT
#define BZCAT "bzcat"
#endif

static int debug = 0;
struct match *match, *top;
static int bz2_decompression;

byte_order_t native_byte_order;

#define min(x,y) ((x) < (y) ? (x) : (y))

int match_count;
int line = 0;

struct regexp_match {
    char *regexp;
    char *message;
    int is_compiled;
    regex_t compiled;
};

struct regexp_match fallbacks[] = {
    {"/(\\*)|(#define)|(void)|(double)|(extern)|(float)|(real)|(struct)|(union)", "C program text", 0},
    {"[Tt]he",				"English text", 0},
    {"(.FLAGS.*=)|(clean:)|(install:)", "make commands text", 0},
    {"\\.((section)|(version)|(file)|(type))",
     "assembler program text", 0},
    {"\\(output",				"Pascal program text", 0},
    {"(Return-Path:)|(To:)|(Received:)",		"mail text", 0},
    {"(Followup-To:)|(Path:)|(NNTP-Posting-Host:)|(Approved:)", "news text", 0},
};

#define NUM_FALLBACKS (sizeof(fallbacks)/sizeof(struct regexp_match))

static struct match compress_match = {
    0,			/* level */
    0,			/* not indirect */
    -1,-1,-1,		/* not indirect --> index fields not used*/
    0,			/* offset */
    STRING,		/* data_type */
    LITTLE_END,		/* byte_order: not used for string */
    EQUAL,		/* compare string for equality */
    2,			/* numbytes */
    "\037\235",		/* compare_string */
    0,			/* compare_num: unused */
    ~0,			/* compare_mask: unused */
    "compress'd data"   /* format_string. */
};

static struct match gzip_match = {
    0,			/* level */
    0,			/* not indirect */
    -1,-1,-1,		/* not indirect --> index fields not used*/
    0,			/* offset */
    STRING,		/* data_type */
    LITTLE_END,		/* byte_order: not used for string */
    EQUAL,		/* compare string for equality */
    2,			/* numbytes */
    "\037\213",		/* compare_string */
    0,			/* compare_num: unused */
    ~0,			/* compare_mask: unused */
    "gzip compressed data"   /* format_string. */
};

static struct match bunzip2_match = {
    0,			/* level */
    0,			/* not indirect */
    -1,-1,-1,		/* not indirect --> index fields not used*/
    0,			/* offset */
    STRING,		/* data_type */
    LITTLE_END,		/* byte_order: not used for string */
    EQUAL,		/* compare string for equality */
    3,			/* numbytes */
    "BZh",		/* compare_string */
    0,			/* compare_num: unused */
    ~0,			/* compare_mask: unused */
    "bunzip2 huffman compressed data"   /* format_string. */
};

void stop_here(void) { }

static void
show_match (char *preamble, struct match *match) {
    switch (match->data_type) {
    case STRING:
	fprintf (stderr,
		 "%s match 0x%p, offset %ld, format_string %s.\n",
		 preamble, match, match->offset, match->format_string);
	break;
    default:
	fprintf (stderr,
		 "%s match 0x%p, offset %ld, byte_order %d numbytes %d num 0x%lx format %s\n",
		 preamble, match, match->offset, match->byte_order,
		 match->numbytes, match->compare_num, match->format_string);
	break;
    }
}

static void
alloc_match(int level) {
    struct match *result;

    match_count++;
    if (match == NULL) {
	if (level != 0) {
	    fprintf (stderr, "file: magic file does not begin with a level zero rule, exiting.\n");
	    exit(1);
	}
    } else {
	while (level < match->level) match = match->parent;
    }
    if (!(result  = malloc(sizeof(struct match)))) {
	fprintf (stderr, "alloc_magic out of memory.  Exiting\n");
	exit(1);
    }
    result->level = level;
    result->sibling = result->child = NULL;
    if (match) {
	if (match->level == level) {
	    match->sibling = result;
	    result->parent = match->parent;
	}
	else if (match->level == level - 1) {
	    match->child = result;
	    result->parent = match;
	} else {
	    fprintf (stderr, "Rule level mismatch: previous rule was level %d, this rule is level %d, line %d, match_count %d.\n",
		     match->level, result->level, line, match_count);
	    exit(2);
	}
    } else {
	result->parent = NULL;
	top = result;
    }
    match = result;
}

static int
parsenum (char *str) {
    int result = 0;
    if (str[0] != '0') sscanf(str, "%d", &result);
    else if (str[1] != 'x') sscanf(str+1, "%o", &result);
    else sscanf (str+2, "%x", &result);
    return result;
}

static char *
getlevel (char *str ,int *level) {
    *level = 0;
    while (*str == '>') {
	(*level)++;
	str++;
    }
    return str;
}

static char *
get_byte_order(char *str, byte_order_t *bo) {
    if (str[1] == 'e') {
	if(str[0] == 'b')
	    *bo = BIG_END;
	else if (str[0] == 'l')
	    *bo = LITTLE_END;
	else
	    goto nomatch;
	return str + 2;
    }
  nomatch:
    *bo = native_byte_order;
    return str;
}

static char *
skip_space(char *str) {
    while (*str == ' ' || *str == '\t' ) str++;
    return str;
}

static char *
skip_num(char *str) {
    if (str[0] == '0') {
	if (str[1] == 'x' || str[1] == 'X') {
	    str += 2;
	    while((*str >= '0' && *str <= '9') ||
		  (*str >= 'a' && *str <= 'f') ||
		  (*str >= 'A' && *str <= 'F')) {
		str++;
	    }
	}
	else {
	    while (*str >= '0' && *str <= '7') str++;
	}
    } else
	while (*str >= '0' && *str <= '9') str++;
    return str;
}

static char*
parse_c_string (char *to, unsigned int *numbytes, char *from,
		int allow_spaces) {
    char *to_orig = to;
    for (;*from && *from != '\n' &&
	      ((*from != '\t' && *from != ' ' )|| allow_spaces); to++) {
	if (*from != '\\') {
	    *to = *(from++);
	    continue;
	}
	from++;
	if (*from == 'x' || *from == 'X') {
	    int charval = 0;
	    int i;
	    for(++from, i = 0; i < 2; i++) {
		if (*from >= '0' && *from <= '9') {
		    charval = charval * 16 + (*from - '0');
		    from++;
		} else if (*from >= 'a' && *from <= 'f') {
		    charval = charval * 16 + (*from - 'a') + 10;
		    from++;
		} else if (*from >= 'A' && *from <= 'F') {
		    charval = charval * 16 + (*from - 'A') + 10;
		    from++;
		} else
		    break;
	    }
	    *to = charval;
	    continue;
	}
	if (*from >= '0' && *from <= '7') {
	    int charval = 0;
	    int i;
	    for(i = 0; i < 3 && *from >= '0' && *from <= '7'; i++) {
		charval = charval * 8 + (*from - '0');
		from++;
	    }
	    *to = charval;
	    continue;
	}
	switch (*(from++)) {
	case 'n': *to = '\n'; break;
	case 't': *to = '\t'; break;
	case 'r': *to = '\r'; break;
	case 'b': *to = '\b'; break;
	case ' ':
	case '>':
	case '<':
	case '=':
	case '&':
	case '^':
	case '#':
	case '\t':
	case '\\':
	    *to = *(from-1);
	    break;
	default:
	    fprintf (stderr, "parse_c_string: Unrecognized backslash escape %c (ascii %d).\n", *(from-1), *(from-1));
	    break;
	}
    }
    *to = '\0';
    *numbytes = to - to_orig;
    return from;
}

%}
%x match_type numeric_match read_format_string
OCTAL	(0[0-7]*)
DECIMAL ([1-9][0-9]*)
HEX     (0x[0-9a-fA-F]*)
COMPLEMENT	("~"?)
NUM	{COMPLEMENT}({OCTAL}|{DECIMAL}|{HEX})
CCHAR	([^ \\\n\t]|(\\([ bntr&^<>#=\\]|([0-7]+)))|(\\x[0-9a-fA-F]([0-9a-fA-F]?)))
CSTRING	({CCHAR}+)
FORMATSTRING	([^\n]*)
LEVEL	([>]*)
SP	([ \t]*)
BYTEORDER	(("le"|"be")?)
DATE	{BYTEORDER}"date"
NUMTYPE		({BYTEORDER}(byte|long|short)("&"{NUM})?)
OP	[!&><^=]?
COMPARE	(({OP}{NUM})|"x")
STRCOMPARE (("!"?{CSTRING})|"x")
MASK	(("&"${NUM})?)
%%
<INITIAL>^{SP}"\n"		{ line++; /* skip blank lines */ }
<INITIAL>^#[^\n]*\n		{ line++; /* skip comments */ }
<INITIAL>^{LEVEL}{SP}{NUM}{MASK}{SP}   {
    				char *ptr;
				int level;

				ptr = getlevel(yytext, &level);
				alloc_match(level);
				ptr = skip_space(ptr);
				match->offset = parsenum(ptr);
				match->indirect = 0;
				ptr= skip_num(ptr);
				if (*ptr == '&') {
				      match->compare_mask = parsenum(ptr+1);
				} else {
				      match->compare_mask = ~0;
				}
    				BEGIN(match_type);
				}
<INITIAL>^{LEVEL}{SP}"("{NUM}(.[bs])?([-+]{NUM})?")"{MASK}{SP}   {
    				char *ptr;
				int sign;
				int level;

				ptr = getlevel(yytext, &level);
				alloc_match(level);
				ptr = skip_space(ptr);
				assert(*ptr == '(');
				ptr++;
				match->index_location = parsenum(ptr);
				ptr = skip_num(ptr);
				if (*ptr == '.') {
				    if (ptr[1] == 'b')
					match->index_size = 1;
				    else
					match->index_size = 2;
				    ptr += 2;
				}
				else
				    match->index_size = 4;
				if (*ptr == '+' || *ptr == '-') {
				  sign = (*ptr == '+') ? 1 : -1;
				  ptr++;
				  match->offset = parsenum(ptr) * sign;
				  ptr= skip_num(ptr);
				} else {
				  match->offset = 0;
				}
				match->indirect = 1;
				assert(*ptr == ')');
				ptr++;
				if (*ptr == '&') {
				      match->compare_mask = parsenum(ptr+1);
				} else {
				      match->compare_mask = ~0;
				}
    				BEGIN(match_type);
				}
<match_type>{NUMTYPE}{SP}	{
    				 char *ptr =
				    get_byte_order(yytext,&match->byte_order);

				    if (!strncmp(ptr, "long", 4)) {
					match->numbytes=4;
				    } else if (!strncmp(ptr, "short", 5)) {
					match->numbytes=2;
				    } else {
					match->numbytes=1;
				    }
				    match->data_type = NUMERIC;
				    BEGIN(numeric_match);
				}
<match_type>{DATE}{SP}		{
				 (void) get_byte_order(yytext,&match->byte_order);
				 match->numbytes = sizeof(time_t);
				 match->data_type = DATE;
				 BEGIN(numeric_match);
				}
<match_type>"string"{SP}("!"?){CSTRING}{SP} {
				    char *ptr;

				    ptr = skip_space(yytext+strlen("string"));
    				    if (yyleng != 0 && yytext[0] == '!') {
					match->comparison = NOT_EQUAL;
					ptr++;
				    } else {
					match->comparison = EQUAL;
				    }
    				    if (!(match->compare_string =
					  malloc(yyleng))) {
					fprintf (stderr, "rest_of_string: out of memory.\n");
					exit(3);
				    }
				    parse_c_string(match->compare_string,
						   &match->numbytes, ptr, 0);
				    match->data_type = STRING;
				    line++;
				    BEGIN(read_format_string);
				    }

<numeric_match>{COMPARE}{SP}	{
    			char *ptr = yytext;
			match->compare_string = 0;
			if (*ptr == 'x' ) {
			    match->comparison = ALWAYS_TRUE;
			    ptr++;
			} else {
			    switch (*ptr) {
			    case '!': match->comparison = NOT_EQUAL; break;
			    case '&': match->comparison = BITS_SET;  break;
			    case '>': match->comparison = GREATER;   break;
			    case '<': match->comparison = LESS;      break;
			    case '^': match->comparison = BITS_CLEAR;break;
			    case '=': match->comparison = EQUAL;     break;
			    default: match->comparison = EQUAL; ptr--;break;
			    }
			    ptr++;
			    match->compare_num = parsenum(ptr);
			    ptr = skip_num(ptr);
			}
			ptr = skip_space(ptr);
			BEGIN(read_format_string);
			}

<read_format_string>{FORMATSTRING}"\n"	{
                        int format_str_len;
			match->format_string = malloc(yyleng);
			if (!match->format_string) {
			    fprintf (stderr, "parsemagic.l: malloc failed.\n");
			    exit(2);
			}
			parse_c_string(match->format_string,
				       &format_str_len, yytext, 1);
			match->format_string[format_str_len] = '\0';
			line++;
			BEGIN(INITIAL);
		    }
.				{printf("unrecognized: %s\n", yytext);}
%%

static char *magic_file = NULL;
static FILE *file_list_file = NULL;
static int follow_symlinks = 0;

static char *file_buf = NULL;
static int buf_size = 0;
static int buf_start = 0;
char *file_name;
static int file_fd;
static struct stat stat_buf;
static int is_piped = 0;
static int last_offset_read = 0;
static int examine_compressed = 0;
static int eof_found = -1;	/* -1 == EOF not yet found. */

static void
usage(void) {
    fprintf (stderr,
	     "Usage: file [options] file1 [file2 file3...]\n"
	     "       file -f indirect_file [options] [file1 file2...]\n"
	     "\n"
	     "Options:\n"
	     "    -f file  Read a newline separated list of files to be checked from file.\n"
	     "             These files are in addition any listed on the command line.\n"
	     "    -L       Follow symbolic links.\n"
	     "    -m file  Read identification numbers from file instead of /etc/magic.\n"
	     "    -v       Print program version and exit.\n"
	     "    -z       Examine the uncompressed contents of compressed files.\n"
	     "\n"
	     " \"-\" can be used in place of any filename to indicate standard input.\n"
	);
}

static int
reset_pipe(void) {
    static int child_pid = 0;
    int pipes[2] = {-1,-1};

    if (child_pid != 0) {
	int status;
	kill(child_pid, SIGKILL);
	if (waitpid(child_pid, &status, 0)<0) {
	    perror("waitpid");
	}
	close(pipes[0]);
	close(pipes[1]);
    }

    last_offset_read = 0;
    pipe(pipes);
    child_pid = fork();
    switch (child_pid) {
    case -1: perror("fork"); return 0;
    case 0: break;
    default: file_fd = pipes[0]; close(pipes[1]); return 1;
    }

    /* Child process  */
    dup2(pipes[1], 1);		/* standard out --> pipe[1] */
    close(pipes[0]);
    close(pipes[1]);
    if (bz2_decompression) {
    	execlp(BZCAT, BZCAT, file_name, 0);
    } else {
    	execlp(ZCAT, ZCAT, file_name, 0);
    }
    perror("execl");
    exit(1);
}

static int
read_file_bytes (char *ptr, int offset, int len) {
    /* Don't allow reading beyond end of file */
    int bytes_left;

    if ((S_ISREG(stat_buf.st_mode) && offset >= stat_buf.st_size) || 
	((eof_found != -1) && offset >= eof_found)) return 0;

    if (is_piped) {
	if (offset < last_offset_read) reset_pipe();
	while(last_offset_read < offset) {
	    char garbage[4096];
	    int bytes_read;
	    int bytes_to_read;
	    bytes_to_read = min(sizeof(garbage),offset - last_offset_read); 
	    if ((bytes_read = read(file_fd, garbage, bytes_to_read)) <= 0) {
		return 0;		/* Attempt to read past EOF, presumably. */
	    }
	    last_offset_read += bytes_read;
	}
    }
    else if (offset != last_offset_read &&
	     lseek(file_fd, offset, SEEK_SET) < 0) {
	perror("lseek");
	return 0;
    }
    bytes_left = len;
    while (bytes_left > 0) {
	int this_count;

	this_count = read(file_fd, ptr, bytes_left);
	if (this_count <= 0) {
	    if (this_count < 0) {
		perror("read");
	    } else {
		eof_found = last_offset_read;
	    }
	    return len - bytes_left;
	}
	bytes_left -= this_count;
	ptr += this_count;
	last_offset_read += this_count;
    }
  
    return len;
}

#define BLOCK_SIZE 1024

static char*
access_file(int offset, int len) {
    char *new_buf = NULL;
    int bytes_needed;
    int bytes_read;

    if ( offset >= buf_start && offset + len <= buf_start + buf_size)
	return file_buf + (offset - buf_start);

    if (offset < buf_start || !file_buf || !(new_buf = realloc(file_buf,offset + len - buf_start))) {
	if (file_buf) free(file_buf);
	buf_start = is_piped ? 0 : offset;
	buf_size = 0;
	if (!(file_buf = malloc(offset + len - buf_start))) {
	    fprintf(stderr, "access_file: out of memory (needed %d bytes).\n",
		    len);
	    return NULL;
	}
	/* Fall through */
    }
    
    if (new_buf) file_buf = new_buf;
    bytes_needed = (offset + len) - (buf_start + buf_size);
    bytes_read = read_file_bytes(file_buf + buf_size, buf_start + buf_size,
				 bytes_needed);
    buf_size += bytes_read;
    if (bytes_read != bytes_needed) {
	return NULL;
    }
    return file_buf + offset - buf_start;
}

static unsigned long
get_value (char *ptr, unsigned int numbytes, byte_order_t byte_order) {
    unsigned long result;
    switch(numbytes) {
    case 1: result = *(unsigned char*)ptr; break;
    case 2: result = *(unsigned short*)ptr; break;
    case 4: result = *(unsigned long*)ptr; break;
    default:
	fprintf (stderr, "get_value: unrecognized size %d.\n", numbytes);
	result = *(unsigned long*)ptr;
	break;
    }
    if (byte_order != native_byte_order) {
	switch(numbytes) {
	case 2: result = ((result&0xFF00)>>8) | ((result&0xFF)<<8);
		break;
	case 4: result = 
		    ((result&0xFF000000)>>24) | ((result&0xFF0000)>>8) |
		    ((result&0xFF00)<<8) | ((result&0xFF)<<24);
		break;
	default: break;
	}
    }
    return result;
}

/* execute_test returns 0 if comparison failes,
   non-zero if comparison succeeds. */
static inline int
execute_test(struct match *match, char *ptr) {
    if (match->comparison == ALWAYS_TRUE) return 1;
    if (match->data_type != STRING) {
	unsigned long value = get_value(ptr, match->numbytes,
					match->byte_order) &
	    match->compare_mask;
	switch (match->comparison) {
	case EQUAL: return value == match->compare_num;
	case NOT_EQUAL: return value != match->compare_num;
	case GREATER: return value > match->compare_num;
	case LESS: return value < match->compare_num;
	case BITS_SET: return value & match->compare_num;
	case BITS_CLEAR: return (~value) & match->compare_num;
	default:
	    fprintf (stderr, "Unknown match comparison %d.\n", match->comparison);
	    return 0; 
	}
    }
    else {
	int compare_result = strncmp(ptr, match->compare_string, match->numbytes);
	switch (match->comparison) {
	case EQUAL: return compare_result == 0;
	case NOT_EQUAL: return compare_result != 0;
	case GREATER: return compare_result > 0;
	case LESS: return compare_result < 0;
	default:
	    fprintf (stderr, "Unknown string match comparison %d.\n", match->comparison);
	    return 0; 
	}
    }
    /*NOTREACHED*/
}


/* test_match:
   If match succeeds, write descriptive text to match_buf and return non-zero.
   If match failes, return zero and leave contents of match_buf unchanged. */
static int
test_match(struct match *match, char *match_buf) {
    char *ptr;
    unsigned long offset;
    char *format_string;

    offset = match->offset;
    if (match->indirect) {
	ptr = access_file(match->index_location, match->index_size);
	if (!ptr) return 0;
	offset += get_value(ptr, match->index_size, match->byte_order);
    }
    ptr = access_file(offset, match->numbytes);
    if(!ptr) return 0;
    if (!execute_test(match,ptr)) {
        if (debug > 1) show_match("did not match",match);
        return 0;	/* If the entry does not match, return zero. */
    }
    if (debug > 0) show_match("matched",match);

    /* This is a correct match.  The rest of this routine is concerned
       with filling in match_buf. */

    format_string = match->format_string;
    if (match->level != 0) {
	if (format_string[0] == '\b')
	    format_string++;
	else
	    *(match_buf++) = ' ';
    }
    switch (match->data_type) {
	case STRING: {
	    char *string;
	    if ((string = malloc(match->numbytes+1)) == NULL) {
		fprintf (stderr, "test_match: Out of memory, needed %d bytes.\n",
			 match->numbytes+1);
		break;
	    }
	    strncpy(string, ptr, match->numbytes);
	    string[match->numbytes] = '\0';
	    sprintf (match_buf, format_string, string);
	    free(string);
	    break;
	}
	case NUMERIC: {
	    unsigned long value = get_value(ptr, match->numbytes,
					    match->byte_order);
	    sprintf (match_buf, format_string, value);
	    break;
	}
	case DATE: {
	    char date_string[50];
	    time_t value = (time_t) get_value(ptr, match->numbytes,
					    match->byte_order);

	    strcpy(date_string, ctime(&value));
	    date_string[24] = '\0';
	    sprintf (match_buf, format_string, date_string);
	    break;
	}
	default:
	    fprintf (stderr, "test_match: Unknown data_type %d.\n",
		     match->data_type);
	    break;
    }
    return 1;
}

static int
do_matches(struct match *match, char *match_buf) {
    int return_value;
    if (!match) return 0;
    if ((return_value = test_match(match, match_buf))) {
	match_buf += strlen(match_buf);
	do_matches(match->child, match_buf);
	if (match->level == 0) {
	    return 1;
	}
    }
    if (do_matches(match->sibling, match_buf))
	return 1;
    else
	return return_value;
}

static void
process_regular_file(char *match_buf) {
    int saved_file_fd = -1;
    static int is_compressed;
    buf_size = 0;
    buf_start = 0;
    is_piped = 0;
    eof_found = -1;

    if (examine_compressed &&
	(test_match(&compress_match, match_buf) ||
	 test_match(&gzip_match, match_buf) ||
	 test_match(&bunzip2_match, match_buf))) {
	bz2_decompression = test_match(&bunzip2_match, match_buf);
	match_buf[0] = '\0';
	saved_file_fd = file_fd;
	is_piped = 1;
	is_compressed = 1;
	reset_pipe();
	if (file_buf) free(file_buf);
	file_buf = NULL;
	buf_size = 0;
	buf_start = 0;
	/* fall through */
    }
    else {
	is_compressed = 0;
    }
    if (!do_matches(top, match_buf)) {
	int i;
	const int fallback_len = 1024;
	char *ptr = access_file(0, fallback_len);
	if (!ptr) ptr = access_file(0, buf_size);
	ptr[buf_size-1] = '\0';	/* Avoid segmentation fault if there are no null bytes. */
	/* STUB: Other methods of determining file type. */
	for (i = 0; ptr && (i < NUM_FALLBACKS); i++ ) {
	    if(!fallbacks[i].is_compiled) {
		if(regcomp(&fallbacks[i].compiled,fallbacks[i].regexp,REG_EXTENDED)) {
		    fprintf (stderr, "regcomp on %d failed (skipping).\n", i);
		    continue;
		}
		fallbacks[i].is_compiled = 1;
	    }
	    if (regexec(&fallbacks[i].compiled, ptr, 0, NULL, 0) == 0) {
		strcpy(match_buf, fallbacks[i].message);
		break;
	    }
	}
	if (i == NUM_FALLBACKS) {
	    int j;
	    if (buf_size == 0) {
		strcpy(match_buf, "empty");
	    }
	    else {
		strcpy(match_buf, "ASCII text");
		for (j = 0; j < buf_size-1; j++) {
		    if (file_buf[j] < '\n' || file_buf[j] >= 127) {
			strcpy(match_buf, "data");
			break;
		    }
		}
	    }
	}
    }
    if (file_fd != 0) close(file_fd);
    if (is_compressed) {
	file_fd = saved_file_fd;
	match_buf += strlen(match_buf);
	*(match_buf++) = ' ';
	*(match_buf++) = '(';
	examine_compressed = 0;
	process_regular_file(match_buf);
	strcat(match_buf, ")");
    }
}

static void
process_file(char *filename) {
    char match_buf[300];

    match_buf[0] = '\0';

    file_name = filename;
    if (strcmp(filename, "-") == 0) {
	file_name = "standard input";
	file_fd = 0;
	fstat(0, &stat_buf);
	process_regular_file(match_buf);
	printf ("standard input: %s\n", match_buf);
	return;
    }

    if ((follow_symlinks ? stat : lstat)(filename, &stat_buf) < 0) {
	perror(filename);
	return;
    }
    if (S_ISREG(stat_buf.st_mode)) {
	if ((file_fd = open(filename, O_RDONLY)) < 0) {
	    perror(filename);
	    return;
	}
	process_regular_file(match_buf);
    }
    else if (S_ISLNK(stat_buf.st_mode)) {
	char linkval[MAXPATHLEN+1];
	int linksize;
	struct stat dummy_stat;
	if ((linksize = readlink(filename,linkval, sizeof(linkval)-1)) < 0) {
	    perror(filename);
	    return;
	}
	linkval[linksize] = '\0';
	sprintf (match_buf, "%ssymbolic link to %s",
		stat(filename, &dummy_stat) < 0 ? "broken " : "",
		linkval);
    }
    else if (S_ISDIR(stat_buf.st_mode)) {
	strcpy(match_buf, "directory");
    }
    else if (S_ISCHR(stat_buf.st_mode) || S_ISBLK(stat_buf.st_mode)) {
	sprintf (match_buf, "%s special (%d/%d)",
		S_ISCHR(stat_buf.st_mode) ? "character" : "block",
		major(stat_buf.st_rdev), minor(stat_buf.st_rdev));
    }
    else if (S_ISFIFO(stat_buf.st_mode)) {
	strcpy (match_buf, "fifo");
    }
    else if (S_ISSOCK(stat_buf.st_mode)) {
	strcpy (match_buf, "socket");
    }
    printf ("%s: %s\n", file_name, match_buf);
}

static FILE*
fopen_read_only(const char *filename) {
    if (strcmp(filename, "-") == 0) {
	return stdin;
    }
    else {
	return fopen(filename, "r");
    }
}

static struct option options[] = {
	{ "check",	0, NULL, 'c' },
	{ "debug",	0, NULL, 'd' },
	{ "from",	1, NULL, 'f' },
	{ "help",	0, NULL, 'h' },
	{ "links",	0, NULL, 'L' },
	{ "magic",	1, NULL, 'm' },
	{ "version",	0, NULL, 'v' },
	{ "compressed",	0, NULL, 'z' },
	{ NULL,		0, NULL,  0  }
};

int
main (int argc, char **argv) {
    int opt;
    const unsigned long byte_test = 0x01020304;

    native_byte_order = (*((char*)&byte_test) == 1) ? BIG_END : LITTLE_END;
    magic_file = getenv("MAGIC");

    while((opt = getopt_long(argc, argv, "cdf:hLm:vz", options, NULL)) != EOF) {
	switch(opt) {
        case 'd':
	    fprintf(stderr, "%s: debug level: %d\n", argv[0], ++debug);
	    break;
	case 'c':
	    break;
	case 'f':
	    if (!(file_list_file = fopen_read_only(optarg))) {
		fprintf(stderr, "%s: -f %s: %s\n",
			argv[0], optarg, strerror(errno));
		continue;
	    }
	    break;
	case 'h':
	    usage();
	    return 0;
	case 'L':
	    follow_symlinks = 1;
	    break;
	case 'm':
	    magic_file = optarg;
	    break;
	case 'v':
	    fprintf (stderr, "%s: freefile version 1.8\n", argv[0]);
	    return 0;
	case 'z':
	    examine_compressed = 1;
	    break;
	default:
	    fprintf (stderr, "%s: unrecognized option \'%c\'.\n", argv[0], opt);
	    break;
	}
    }

    if(magic_file == NULL) {
	int	e1, e2;

	if ((yyin = fopen_read_only("/usr/share/magic")) == NULL) {
	    e1 = errno;
	    yyin = fopen_read_only("/etc/magic");
	    e2 = errno;
	}
	if(yyin == NULL) {
	    fprintf(stderr, "%s: /usr/share/magic: %s\n",
			argv[0], strerror(e1));
	    fprintf(stderr, "%s: /etc/magic: %s\n",
			argv[0], strerror(e2));
	    return 1;
	}
    } else if (!(yyin = fopen_read_only(magic_file))) {
	fprintf(stderr, "%s: %s: %s\n", argv[0], magic_file, strerror(errno));
	return 1;
    }

    yylex();
    if (optind == argc && !file_list_file) {
	usage();
	return 1;
    }

    while(optind < argc) {
	process_file(argv[optind]);
	optind++;
    }
    if(file_list_file) {
	char filename[MAXPATHLEN];
	while (fgets(filename, sizeof(filename), file_list_file)) {
	    filename[strlen(filename)-1] = '\0'; /*remove trailing linefeed.*/
	    process_file(filename);
	}
    }
    return 0;
}
